// Author 		: F. de la Parra
// Date			: April, 2019
// Description 	: The MTOptimizer class implements the main flow of the 
//				  program to read in a daily riderhsip file containing 
//				  records of public transportation usage in the city of
//				  Toronto. This class also generates, and stores in a file
//				  named "InOperationFleets.txt", optimal hourly fleet 
//				  configurations of vehicles of a given type (trains, buses,
//				  etc.). Finally, file it reports any bad lines, identified
//				  when reading in file "ridership.txt", in file "errorlog.txt".
// Version		: 1.0

import java.util.Scanner;
import java.io.FileInputStream;
import java.io.PrintWriter;
import java.io.FileOutputStream;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.StringTokenizer;
import java.util.Collections;

public class MTOptimizer {
	// Scanner objects to read the vehicle type files.
	Scanner subwaysFile = null;
	Scanner busesFile = null;
	Scanner streetcarsFile = null;
	Scanner gobusesFile = null;
	Scanner gotrainsFile = null;
	
	// Array lists to keep in memory data about subways,
	// buses, streetcars, GObuses and GOtrains.
	ArrayList<Subway> subways = new ArrayList<Subway>();
	ArrayList<Bus> buses = new ArrayList<Bus>();
	ArrayList<Streetcar> streetcars = new ArrayList<Streetcar>();
	ArrayList<GObus> gobuses = new ArrayList<GObus>();
	ArrayList<GOtrain> gotrains = new ArrayList<GOtrain>();
	
	// Attributes and objects to process the "readership.txt" files.
	ArrayList<Rider> riders = new ArrayList<Rider>(65000);
	Scanner ridership = null;
	String line = null;
	int lineNumber = 0;
		
	// Variables to create lines of the "errorlog.txt" file.
	PrintWriter errorlog = null;
	boolean firstErrorMsg = true;
	String errorMsg = null;
	String errorLine = null;
	
	// Reference to "InOperationFleets.txt" file.
	PrintWriter inOperationFleets = null;
	
	// Method "main" drives the stages for finding hourly optimal vehicle fleets.
	public static void main(String[] args) {
		
		// Create a program object.
		MTOptimizer MTOptimizerObject = new MTOptimizer();
		
		// Read "ridership.txt" file into the "riders" array list. Skip and report
		// all bad lines in file "errorlog.txt".
		MTOptimizerObject.readRidership();
		
		// Read all vehicle files into their respective array lists.
		MTOptimizerObject.readVehicles();
		
		// Find hourly optimal vehicle fleets and generate "InOperationFleets.txt" file. 
		MTOptimizerObject.findOptimalFleets();
	}
	
	// This method reads the "ridership.txt" file into the "riders" array list.
	private void readRidership() {
		
		// Open for 'read' the "ridership.txt" file.
		try {
			ridership = new Scanner(new FileInputStream("ridership.txt"));
		}
		catch (FileNotFoundException e) {	
			System.out.println("Error opening the file 'ridership.txt'. MTOptimizer aborted!");
			System.exit(0);
		}

		// Create and open for 'write' the "errolog.txt" file.
		try {
			errorlog = new PrintWriter(new FileOutputStream("errorlog.txt"));
		}
		catch (FileNotFoundException e) {	
			System.out.println("Error opening the file 'errorlog.txt'. MTOptimizer aborted!");
			System.exit(0);
		}
		
		// Read the "ridership.txt" file into the "riders" array list

		StringTokenizer tokens = null;
		String[] values = new String[5];
		String value = null;
		int valueIndex;					// Index to values in a ridership line.
		boolean noError;
		
		// Read the "ridership.txt" file, line by line, into the "riders" array list.
		// Send any invalid lines in the "ridership.txt" to the "errorlog.txt" file.
		while (ridership.hasNextLine()) {
			line = ridership.nextLine();
			lineNumber += 1;
			tokens = new StringTokenizer(line, ",");
			noError = true;
			if (tokens.countTokens() == 5) {
				valueIndex = 0;
				while (tokens.hasMoreTokens()) {
					value = tokens.nextToken();
					if (checkValue(value, valueIndex, noError))
						values[valueIndex] = value;
					else 
						noError = false;
					valueIndex += 1;					
				}
				if (noError)
					riders.add(new Rider(values[0], values[1], values[2],
										 Integer.parseInt(values[3]), values[4])); 
			}
			else
				irregularLine(tokens.countTokens());
		}
		
		// Close files to flush all data remaining in temporary I/O buffers.
		ridership.close();
		errorlog.close();
	}
	
	// This method reads the vehicle files into their respective array lists.
	private void readVehicles() {
		StringTokenizer tokens = null;
		String value = null;
		int valueIndex;					// Index to values in vehicle type files.
		
		// Open for 'read' the "subways.txt" file.
		try {
			subwaysFile = new Scanner(new FileInputStream("subways.txt"));
		}
		catch (FileNotFoundException e) {	
			System.out.println("Error opening the file 'subways.txt'. MTOptimizer aborted!");
			System.exit(0);
		}
		
		// Open for 'read' the "gotrains.txt" file.
		try {
			gotrainsFile = new Scanner(new FileInputStream("gotrains.txt"));
		}
		catch (FileNotFoundException e) {	
			System.out.println("Error opening the file 'gotrains.txt'. MTOptimizer aborted!");
			System.exit(0);
		}
		
		// Open for 'read' the "streetcars.txt" file.
		try {
			streetcarsFile = new Scanner(new FileInputStream("streetcars.txt"));
		}
		catch (FileNotFoundException e) {	
			System.out.println("Error opening the file 'streetcars.txt'. MTOptimizer aborted!");
			System.exit(0);
		}
		
		// Open for 'read' the "buses.txt" file.
		try {
			busesFile = new Scanner(new FileInputStream("buses.txt"));
		}
		catch (FileNotFoundException e) {	
			System.out.println("Error opening the file 'buses.txt'. MTOptimizer aborted!");
			System.exit(0);
		}
		
		// Open for 'read' the "gobuses.txt" file.
		try {
			gobusesFile = new Scanner(new FileInputStream("gobuses.txt"));
		}
		catch (FileNotFoundException e) {	
			System.out.println("Error opening the file 'gobuses.txt'. MTOptimizer aborted!");
			System.exit(0);
		}
		
		// Read the "subways.txt" file into the "subways" array list.
		String[] values = new String[6];
		while (subwaysFile.hasNextLine()) {
			line = subwaysFile.nextLine();
			tokens = new StringTokenizer(line, ",");
			for (int i = 0; i < 6; i++)
				values[i] = tokens.nextToken();
			subways.add(new Subway(values[0], values[1], Integer.parseInt(values[2]),
								   Integer.parseInt(values[3]), values[4], values[5]));
		}

		// Read the "gotrains.txt" file into the "gotrains" array list.		
		values = new String[3];
		while (gotrainsFile.hasNextLine()) {
			line = gotrainsFile.nextLine();
			tokens = new StringTokenizer(line, ",");
			for (int i = 0; i < 3; i++)
				values[i] = tokens.nextToken();
			gotrains.add(new GOtrain(values[0], values[1], Integer.parseInt(values[2])));
		}

		// Read the "streetcars.txt" file into the "streetcarss" array list.		
		values = new String[3];
		while (streetcarsFile.hasNextLine()) {
			line = streetcarsFile.nextLine();
			tokens = new StringTokenizer(line, ",");
			for (int i = 0; i < 3; i++)
				values[i] = tokens.nextToken();
			streetcars.add(new Streetcar(values[0], values[1], values[2]));
		}

		// Read the "buses.txt" file into the "buses" array list.		
		values = new String[3];
		while (busesFile.hasNextLine()) {
			line = busesFile.nextLine();
			tokens = new StringTokenizer(line, ",");
			for (int i = 0; i < 3; i++)
				values[i] = tokens.nextToken();
			buses.add(new Bus(values[0], values[1], Integer.parseInt(values[2])));
		}
			
		// Read the "gobuses.txt" file into the "gobuses" array list.		
		values = new String[3];
		while (gobusesFile.hasNextLine()) {
			line = gobusesFile.nextLine();
			tokens = new StringTokenizer(line, ",");
			for (int i = 0; i < 3; i++)
				values[i] = tokens.nextToken();
			gobuses.add(new GObus(values[0], values[1], Integer.parseInt(values[2])));
		}
	
		// Close all vehicle files.
		subwaysFile.close();
		gotrainsFile.close();		
		streetcarsFile.close();
		busesFile.close();
		gobusesFile.close();		
	}

	// This method finds hourly optimal vehicle fleets (i.e., minimum number of vehicles
	// to service an hourly passenger load) and reports them in file "InOperationFleets.txt".
	private void findOptimalFleets() {
		double load = 0;
		HourlyLoad[] hl = new HourlyLoad[24];
		for(int i = 0; i < 24; i++)
			hl[i] = new HourlyLoad();
		
		for(Rider item : riders) {
			if (item.ageGroup.equals("C"))
				load = 0.75;
			else if (item.ageGroup.equals("A"))
				load = 1;
			else if (item.ageGroup.equals("S"))
				load = 1.25;
			
			if (item.transportType.equals("S"))
				hl[item.hour - 1].incSubwayLoad(load);
			else if (item.transportType.equals("G"))
				hl[item.hour - 1].incGOtrainLoad(load);
			else if (item.transportType.equals("X"))
				hl[item.hour - 1].incStreetcarLoad(load);
			else if (item.transportType.equals("C"))
				hl[item.hour - 1].incBusLoad(load);
			else if (item.transportType.equals("D"))
				hl[item.hour - 1].incGObusLoad(load);			
		}
		
		// Create and open for 'write' the "InOperationFleets.txt" file.
		try {
			inOperationFleets = new PrintWriter(new FileOutputStream("InOperationFleets.txt"));
		}
		catch (FileNotFoundException e) {	
			System.out.println("Error opening the file 'InOperationFleets.txt'. MTOptimizer aborted!");
			System.exit(0);
		}
		
		// Generate and write the hourly optimal fleets, for all vehicle types, to
		// the file "InOperationFleets.txt".
		int hourCount = 1;
		for (HourlyLoad item : hl) {
			inOperationFleets.write("[Hour = " + hourCount + "]\n");
			
			inOperationFleets.write("[Subway Fleet]\n");
			genOptimalSubwayFleet(item.getSubwayLoad());
			inOperationFleets.write("\n");
			
			inOperationFleets.write("[GOtrain Fleet]\n");
			genOptimalGotrainFleet(item.getGotrainLoad());
			inOperationFleets.write("\n");
			
			inOperationFleets.write("[Streetcar Fleet]\n");	
			genOptimalStreetcarFleet(item.getStreetcarLoad());
			inOperationFleets.write("\n");
			
			inOperationFleets.write("[Bus Fleet]\n");
			genOptimalBusFleet(item.getBusLoad());
			inOperationFleets.write("\n");
			
			inOperationFleets.write("[GObus Fleet]\n");
			genOptimalGobusFleet(item.getGobusLoad());

			if (hourCount < 24) {
				inOperationFleets.write("\n");
				hourCount += 1;
			}
		}
		
		inOperationFleets.close();
	}
	
	// This method generates the optimal subway fleet for a given hourly passenger load.
	private void genOptimalSubwayFleet(double hourlyLoad) {
		if (hourlyLoad == 0.0)
			inOperationFleets.write("Out of Service\n");
		
		// Sort subways in descending order of passenger capacity.
		descendingSort(subways);
								
		int fleetCapacity = 0;					
		for (Subway s : subways) {
			if (fleetCapacity >= hourlyLoad)
				break;
			if (s.operationalStatus.equals("A")) {
				fleetCapacity += s.getCapacity();	
				inOperationFleets.write("train -> unit number = " + s.getUnitNumber() +
										", " + "id = " + s.getTrainId() + ", " +
										"capacity = " + s.getCapacity() + "\n");
			}
		}
	}
	
	// This method generates the optimal GOtrain fleet for a given hourly passenger load.
	private void genOptimalGotrainFleet(double hourlyLoad) {
		if (hourlyLoad == 0.0)
			inOperationFleets.write("Out of Service\n");
		
		// Sort GOtrains in descending order of passenger capacity.
		descendingSort(gotrains);
								
		int fleetCapacity = 0;					
		for (GOtrain g : gotrains) {
			if (fleetCapacity >= hourlyLoad)
				break;
			fleetCapacity += g.getCapacity();	
			inOperationFleets.write("gotrain -> unit number = " + g.getUnitNumber() + ", " +
									"id = " + g.getTrainId() + ", " +
									"capacity = " + g.getCapacity() + "\n");
		}
	}

	// This method generates the optimal streetcar fleet for a given hourly passenger load.
	private void genOptimalStreetcarFleet(double hourlyLoad) {
		if (hourlyLoad == 0.0)
			inOperationFleets.write("Out of Service\n");
		// Sort streetcars in descending order of passenger capacity.
		descendingSort(streetcars);
								
		int fleetCapacity = 0;					
		for (Streetcar st : streetcars) {
			if (fleetCapacity >= hourlyLoad)
				break;
			fleetCapacity += st.getCapacity();	
			inOperationFleets.write("streetcar -> unit number = " + st.getUnitNumber() + ", " +
									"id = " + st.getStreetcarId() + ", " +
									"capacity = " + st.getCapacity() + "\n");
		}			
	}
	
	// This method generates the optimal bus fleet for a given hourly passenger load.
	private void genOptimalBusFleet(double hourlyLoad) {
		if (hourlyLoad == 0.0)
			inOperationFleets.write("Out of Service\n");
		// Sort subways in descending order of passenger capacity.
		descendingSort(buses);
								
		int fleetCapacity = 0;					
		for (Bus b : buses) {
			if (fleetCapacity >= hourlyLoad)
				break;
			fleetCapacity += b.getCapacity();	
			inOperationFleets.write("bus -> unit number = " + b.getUnitNumber() + ", " +
									"id = " + b.getBusId() + ", " +
									"capacity = " + b.getCapacity() + "\n");
		}		
	}
	
	// This method generates the optimal GObus fleet for a given hourly passenger load.
	private void genOptimalGobusFleet(double hourlyLoad) {
		if (hourlyLoad == 0.0)
			inOperationFleets.write("Out of Service\n");
		// Sort subways in descending order of passenger capacity.
		descendingSort(gobuses);
								
		int fleetCapacity = 0;					
		for (GObus gb : gobuses) {
			if (fleetCapacity >= hourlyLoad)
				break;
			fleetCapacity += gb.getCapacity();	
			inOperationFleets.write("gobus -> unit number = " + gb.getUnitNumber() + ", " +
									"id = " + gb.getGobusId() + ", " +
									"capacity = " + gb.getCapacity() + "\n");
		}			
	}
	
	// Auxiliary methods.
	
	// Sort an array list of vehicles in decending order of capacity.
	public void descendingSort(ArrayList<? extends Vehicle> list) {
		Collections.sort(list, (Vehicle v1, Vehicle v2) -> {
							if (v1.getCapacity() > v2.getCapacity())
								return -1;
							else if (v1.getCapacity() < v2.getCapacity())
								return 1;
							else
								return 0;
						});
	}
	
	// This method checks for legal values in a ridership line.
	boolean checkValue(String value, int i, boolean noError) {
		boolean valid = true;
		String s1 = null, sRest = null;
		
		switch (i) {
			case 0:
				s1 = value.substring(0,1);
				sRest = value.substring(1);
				if (!(s1.equals("T") || s1.equals("*") | 
					 (isDigit(s1) && isNumber(sRest)))) {
					errorMsg = "\nError: Invalid data in 'identification' field.";
					valid = false;
				}
				break;
			case 1:
				if (!(value.equals("S") || value.equals("G") || 
					 value.equals("X") || value.equals("C") || value.equals("D"))) {
					 errorMsg = "\nError: Invalid vehicle type.";
					 valid = false;
				}
				break;
			case 2:
				if (!(value.equals("C") || value.equals("A") || 
					 value.equals("S"))) {
					 errorMsg = "\nError: Invalid age group.";
					 valid = false;
				}
				break;
			case 3:
				if (!(isNumber(value))) {
					 errorMsg = "\nError: Invalid not numeric hour of day value.";
					 valid = false;
				}
				break;
			case 4:
				if (!(isNumber(value))) {
					 errorMsg = "\nError: Invalid not numeric date.";
					 valid = false;
					 break;
				}
				if (!(value.substring(0, 4).equals("2019"))) {
					 errorMsg = "\nError: Invalid date format.";
					 valid = false;
					 break;
				}
				break;
		}
		
		if (!valid) {	
			if (firstErrorMsg) {
				firstErrorMsg = false;
				errorLine = "Line " + lineNumber + " --> " + line;
			}
			else
				errorLine = "\n\nLine " + lineNumber + " --> " + line;
			
			if (noError)
				errorLine += errorMsg;
			else
				errorLine = errorMsg;
			errorlog.write(errorLine);
		}
		return valid;
	}

	// This method checks whether a single-character string is a digit.
	boolean isDigit(String s) {
		boolean isNum = false;
		String[] digits = {"0", "1", "2", "3", "4", "5", "6", "7", "8", "9"};
		if (s.length() == 1)
			for (int i = 0; i < digits.length; i++)
				if (digits[i].equals(s)) {
					isNum = true;
					break;
				}
		return isNum;
	}
	
	// This method checks whether a string contains only digits.
	boolean isNumber(String s) {
		boolean isNum = true;
		for (int i = 0; i < s.length(); i++)
			if (!isDigit(s.substring(i, i + 1))) {
				isNum = false;
				break;
			}
		return isNum;
	}

	// This method prints error messages for lines shorter or larger than 5 values.
	void irregularLine(int tokensCount){
		if (tokensCount < 5) {
			errorMsg = "\nError: Insufficient number of values in input line.";
		}
		else {
			errorMsg = "\nError: Too many values in input line.";
		}
		
		if (firstErrorMsg) {
			firstErrorMsg = false;
			errorLine = "Line " + lineNumber + " --> " + line + errorMsg;
		}
		else
			errorLine = "\n\nLine " + lineNumber + " --> " + line + errorMsg;
		errorlog.write(errorLine);
	}
}